# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import unittest
from ibm_appconfiguration.configurations.models import Segment


class MyTestCase(unittest.TestCase):
    sut = None

    def set_up(self):
        segment = {
            'name': 'RegionalUser',
            'segment_id': 'kdu77n4s',
            'rules': [
                {
                    'values': [
                        100
                    ],
                    'operator': 'lesserThanEquals',
                    'attribute_name': 'radius'
                },
                {
                    'values': [
                        50
                    ],
                    'operator': 'lesserThan',
                    'attribute_name': 'cityRadius'
                }
            ]
        }

        self.sut = Segment(segment)

    def test_segment(self):
        self.set_up()
        self.assertTrue(self.sut.get_name() == "RegionalUser")
        self.assertTrue(self.sut.get_segment_id() == "kdu77n4s")
        self.assertTrue(len(self.sut.get_rules()) == 2)
        client_attributes = {
            'radius':  100,
            'cityRadius': 35
        }
        self.assertTrue(self.sut.evaluate_rule(client_attributes))



if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os
import time
import responses
from ibm_appconfiguration import Property, Feature
from ibm_appconfiguration.configurations.configuration_handler import ConfigurationHandler
from ibm_appconfiguration.configurations.internal.utils.metering import Metering
from ibm_appconfiguration.configurations.internal.utils.url_builder import URLBuilder


class MyTestCase(unittest.TestCase):

    def setUp(self):
        self.responses = responses.RequestsMock()
        self.responses.start()
        self.addCleanup(self.responses.stop)
        self.addCleanup(self.responses.reset)
        URLBuilder.set_auth_type(False)
        self.sut = ConfigurationHandler.get_instance()
        self.sut.init("region", "guid", "apikey", None)
        FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'user.json')
        options = {
            'persistent_cache_dir': None,
            'bootstrap_file': FILE,
            'live_config_update_enabled': False
        }
        self.sut.set_context("collectionId", "environmentId", options)
        self.sut.load_data()
        time.sleep(2.5)

    def tearDown(self):
        self.sut = None

    def test_load_from_web(self):
        Metering.get_instance().set_repeat_calls(False)
        mock_response = '''
            {
                "features": [
                    {
                        "name": "featurestring",
                        "feature_id": "featurestring",
                        "type": "STRING",
                        "enabled_value": "Hello",
                        "disabled_value": "Hi",
                        "segment_rules": [],
                        "enabled": true
                    }
                ],
                "properties": [
                    {
                        "name": "numericproperty",
                        "property_id": "numericproperty",
                        "tags": "",
                        "type": "NUMERIC",
                        "value": 30,
                        "segment_rules": [],
                        "created_time": "2021-05-23T08:00:56Z",
                        "updated_time": "2021-05-23T08:00:56Z"
                    }
                ],
                "segments": []
            }
        '''
        url = 'https://region.apprapp.cloud.ibm.com/apprapp/feature/v1/instances/guid/collections/collection_id/config?environment_id=environment_id'
        self.responses.add(responses.GET,
                           url,
                           body=mock_response,
                           content_type='application/json',
                           status=200)
        self.sut.init("region", "guid", "apikey", None)
        options = {
            'persistent_cache_dir': None,
            'bootstrap_file': None,
            'live_config_update_enabled': True
        }
        self.sut.set_context("collection_id", "environment_id", options)
        self.sut.load_data()
        features = self.sut.get_features()
        self.assertEqual(len(features), 1)

    def test_evaluate_property(self):
        property_json = {
            "name": "numericProperty",
            "property_id": "numericproperty",
            "description": "testing prop",
            "value": 10,
            "type": "NUMERIC",
            "tags": "test",
            "segment_rules": [
                {
                    "rules": [
                        {
                            "segments": [
                                "keuyclvf"
                            ]
                        }
                    ],
                    "value": 81,
                    "order": 1
                }
            ],
            "collections": [{
                "collection_id": "appcrash"
            }]
        }
        property_obj = Property(property_json)
        value = self.sut.property_evaluation(property_obj, "id1", {"email": "test.dev@tester.com"})
        self.assertEqual(value, 81)

        value = self.sut.property_evaluation(property_obj, "id1", {"email": "test@f.com"})
        self.assertEqual(value, 10)

        value = self.sut.property_evaluation(property_obj, "id1", {})
        self.assertEqual(value, 10)

    def test_evaluate_feature(self):
        feature_json = {
            "name": "defaultFeature",
            "feature_id": "defaultfeature",
            "type": "STRING",
            "enabled_value": "hello",
            "disabled_value": "Bye",
            "segment_rules": [
                {
                    "rules": [
                        {
                            "segments": [
                                "kg92d3wa"
                            ]
                        }
                    ],
                    "value": "Welcome",
                    "order": 1
                }
            ],
            "segment_exists": True,
            "enabled": True
        }
        feature_obj = Feature(feature_json)
        value = self.sut.feature_evaluation(feature_obj, True, "id1", {"email": "test.dev@tester.com"})
        self.assertEqual(value, "Welcome")

        value = self.sut.feature_evaluation(feature_obj, True, "id1", {"email": "test@tester.com"})
        self.assertEqual(value, "hello")

        value = self.sut.feature_evaluation(feature_obj, True, "id1", {})
        self.assertEqual(value, "hello")

    def test_get_methods(self):
        feature = self.sut.get_feature("defaultfeature")
        self.assertEqual(feature.get_feature_id(), "defaultfeature")

        features = self.sut.get_features()
        self.assertEqual(len(features), 3)

        property_obj = self.sut.get_property("numericproperty")
        self.assertEqual(property_obj.get_property_id(), "numericproperty")

        properties = self.sut.get_properties()
        self.assertEqual(len(properties), 1)

    # for both properties and features
    def test_yaml_evaluation(self):
        Metering.get_instance().set_repeat_calls(False)
        mock_response = '''
            {
                "features": [
                    {
                        "name": "yamlFeature",
                        "feature_id": "yamlFeature",
                        "type": "STRING",
                        "format": "YAML",
                        "enabled_value": "value: enabled",
                        "disabled_value": "value: disabled",
                        "segment_rules": [
                            {
                                "rules": [
                                    {
                                        "segments": [
                                            "reqbody"
                                        ]
                                    }
                                ],
                                "value": "value: targeted",
                                "order": 1
                            }
                        ],
                        "enabled": true
                    }
                ],
                "properties": [
                    {
                        "name": "yamlProperty",
                        "property_id": "yamlProperty",
                        "tags": "",
                        "type": "STRING",
                        "format": "YAML",
                        "value": "value: enabled",
                        "segment_rules": [
                            {
                                "rules": [
                                    {
                                        "segments": [
                                            "reqbody"
                                        ]
                                    }
                                ],
                                "value": "value: targeted",
                                "order": 1
                            }
                        
                        ],
                        "created_time": "2021-05-23T08:00:56Z",
                        "updated_time": "2021-05-23T08:00:56Z"
                    }
                ],
                "segments": [
                    {
                        "name": "reqbody",
                        "segment_id": "reqbody",
                        "rules": [
                            {
                                "values": [
                                    "tester.com"
                                ],
                                "operator": "endsWith",
                                "attribute_name": "email"
                            }
                        ]
                    }
                ]
            }
        '''
        url = 'https://region.apprapp.cloud.ibm.com/apprapp/feature/v1/instances/guid/collections/collection_id/config?environment_id=environment_id'
        self.responses.add(responses.GET,
                           url,
                           body=mock_response,
                           content_type='application/json',
                           status=200)
        self.sut.init("region", "guid", "apikey", None)
        options = {
            'persistent_cache_dir': None,
            'bootstrap_file': None,
            'live_config_update_enabled': True
        }
        self.sut.set_context("collection_id", "environment_id", options)
        self.sut.load_data()
        features = self.sut.get_features()
        properties = self.sut.get_properties()
        self.assertEqual(features['yamlFeature'].get_current_value("id1", {"email": "test.dev@tester.com"})['value'],
                         "targeted")
        self.assertEqual(properties['yamlProperty'].get_current_value("id1", {"email": "test.dev@tester.com"})['value'],
                         "targeted")


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines the model of a rule defined for a segment in App Configuration service.
"""

from typing import Tuple


class Rule:
    """
        Attributes:
           rules (dict): rules JSON object that contains all the Rules.
    """

    def __init__(self, rules: {}):
        self.__attribute_name = rules.get("attribute_name", "")
        self.__operator = rules.get("operator", "")
        self.__values = rules.get("values", list())

    def get_attributes(self) -> str:
        """Get the Rule attributes"""
        return self.__attribute_name

    def get_operator(self) -> str:
        """Get the Rule operator"""
        return self.__operator

    def get_values(self) -> list:
        """Get the Rule values"""
        return self.__values

    def __ends_with(self, key, value) -> bool:
        return key.endswith(value)

    def __starts_with(self, key, value) -> bool:
        return key.startswith(value)

    def __contains(self, key, value) -> bool:
        return value in key

    def __is(self, key, value) -> bool:
        if type(key) is type(value):
            return key == value
        if str(key) == str(value):
            return True
        return False

    def __greater_than(self, key, value) -> bool:

        key_obj = self.__number_conversion(key)
        value_obj = self.__number_conversion(value)

        if key_obj[0] and value_obj[0]:
            return key_obj[1] > value_obj[1]
        return False

    def __lesser_than(self, key, value) -> bool:
        key_obj = self.__number_conversion(key)
        value_obj = self.__number_conversion(value)

        if key_obj[0] and value_obj[0]:
            return key_obj[1] < value_obj[1]
        return False

    def __greater_than_equals(self, key, value) -> bool:
        key_obj = self.__number_conversion(key)
        value_obj = self.__number_conversion(value)

        if key_obj[0] and value_obj[0]:
            return key_obj[1] >= value_obj[1]
        return False

    def __lesser_than_equals(self, key, value) -> bool:
        key_obj = self.__number_conversion(key)
        value_obj = self.__number_conversion(value)

        if key_obj[0] and value_obj[0]:
            return key_obj[1] <= value_obj[1]
        return False

    def __operator_check(self, key_data=None, value_data=None) -> bool:
        key = key_data
        value = value_data

        result = False

        if key is None or value is None:
            return result

        case_checker = {
            "endsWith": self.__ends_with,
            "startsWith": self.__starts_with,
            "contains": self.__contains,
            "is": self.__is,
            "greaterThan": self.__greater_than,
            "lesserThan": self.__lesser_than,
            "greaterThanEquals": self.__greater_than_equals,
            "lesserThanEquals": self.__lesser_than_equals
        }
        return case_checker.get(self.__operator, False)(key, value)

    def __number_conversion(self, value) -> Tuple[bool, float]:
        if isinstance(value, bool):
            return False, 0
        if isinstance(float(value), float):
            return True, float(value)
        return False, 0

    def evaluate_rule(self, entity_attributes: dict) -> bool:
        """Evaluate the the Rule

        Args:
            entity_attributes: Entity attributes object
        """
        result = False
        if self.__attribute_name in entity_attributes:
            key = entity_attributes.get(self.__attribute_name)
        else:
            return result
        for i in range(0, len(self.__values)):
            value = self.__values[i]
            if self.__operator_check(key, value):
                result = True

        return result



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file defines the various messages used by the SDK.
"""

REGION_ERROR = "Provide a valid region in App Configuration init"
GUID_ERROR = "Provide a valid guid in App Configuration init"
APIKEY_ERROR = "Provide a valid apiKey in App Configuration init"
COLLECTION_ID_VALUE_ERROR = "Provide a valid collection_id in App Configuration set_context method"
ENVIRONMENT_ID_VALUE_ERROR = "Provide a valid environment_id in App Configuration set_context method"
COLLECTION_INIT_ERROR = "Invalid action in App Configuration. This action can be performed only after a successful " \
                        "initialization operation. Please check the initialization section for errors. "
SET_CONTEXT_OPTIONAL_ARGUMENTS_ERROR = "Invalid optional parameters passed to set_context method. Either wrong arguments keys or invalid value of arguments passed. " \
                                       "The only available arguments are persistent_cache_dir(str), bootstrap_file(str) & live_config_update_enabled(bool)." \
                                       "Please check the sdk initialization section in the readme."
DEPRECATION_WARNING_MESSAGE = "DeprecationWarning: With v0.2.3, the existing method of passing the optional arguments `configuration_file` & `live_config_update_enabled` to set_context method is deprecated and will be removed from v0.3.0 Use the options argument instead."
ERROR_NO_WRITE_PERMISSION = "Persistent cache directory provided does not have write permission. Make sure the directory has required access",
BOOTSTRAP_FILE_NOT_FOUND_ERROR = "Provide a valid bootstrap_file path while live_config_update_enabled is false in set_context method."
CONFIGURATION_HANDLER_INIT_ERROR = 'Invalid action in ConfigurationHandler. This action can be performed only after a ' \
                                   'successful initialization. Please check the initialization section for errors. '
CONFIGURATION_HANDLER_METHOD_ERROR = "Invalid action in ConfigurationHandler. Should be a method/function"
SINGLETON_EXCEPTION = "class must be initialized using the get_instance() method."
FEATURE_INVALID = "Invalid feature_id - "
NO_INTERNET_CONNECTION_ERROR = 'No connection to internet. Please re-connect.'
PROPERTY_INVALID = "Invalid property_id - "
CONFIGURATIONS_FETCH_SUCCESS = "Successfully fetched the configurations."
RETRY_AFTER_TEN_MINUTES = "Failed to fetch the configurations. Retrying after 10 minutes."



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
IBM Cloud App Configuration is a centralized feature management and configuration service on IBM Cloud
for use with web and mobile applications, microservices, and distributed environments.

Instrument your applications with App Configuration Python SDK, and use the App Configuration dashboard,
CLI or API to define feature flags or properties, organized into collections and targeted to segments.
Toggle feature flag states in the cloud to activate or deactivate features in your application or environment, when required.
You can also manage the properties for distributed applications centrally.
"""
from typing import Dict, Optional
from .configurations.internal.utils.validators import Validators
from .configurations.models import Feature, Property
from .configurations.internal.utils.logger import Logger
from .configurations.internal.common import config_messages
from .configurations.configuration_handler import ConfigurationHandler


class AppConfiguration:
    """ AppConfiguration class"""
    __instance = None

    # regions
    REGION_US_SOUTH = "us-south"
    REGION_EU_GB = "eu-gb"
    REGION_AU_SYD = "au-syd"
    REGION_US_EAST = "us-east"
    override_server_host = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if AppConfiguration.__instance is None:
            return AppConfiguration()
        return AppConfiguration.__instance

    @staticmethod
    def enable_debug(enable: bool):
        """Set the logger in debug mode

        Args:
            enable: A boolean value to set the logger debug mode
        """
        Logger.set_debug(enable)

    def __init__(self):
        """ Virtually private constructor. """

        if AppConfiguration.__instance is not None:
            raise Exception("AppConfiguration " + config_messages.SINGLETON_EXCEPTION)

        self.__apikey = ''
        self.__region = ''
        self.__configuration_handler_instance = None
        self.__guid = ''
        self.__is_initialized = False
        self.__is_initialized_configuration = False
        self.__is_loading = False
        AppConfiguration.__instance = self

    def init(self, region: str, guid: str, apikey: str):
        """Initialise the AppConfiguration

           Args:
               region: Region Region name where the service instance is created.
               guid : GUID of the App Configuration service.\
               Get it from the service credentials section of the dashboard
               apikey : ApiKey of the App Configuration service.\
               Get it from the service credentials section of the dashboard
        """

        if not Validators.validate_string(region):
            Logger.error(config_messages.REGION_ERROR)
            return
        if not Validators.validate_string(apikey):
            Logger.error(config_messages.APIKEY_ERROR)
            return
        if not Validators.validate_string(guid):
            Logger.error(config_messages.GUID_ERROR)
            return
        self.__apikey = apikey
        self.__region = region
        self.__guid = guid
        self.__is_initialized = True
        self.__is_loading = False
        self.__setup_configuration_handler()

    def get_region(self) -> str:
        """Get the region currently used by the service.

        Returns:
            The region string currently used by the service.
        """
        return self.__region

    def get_guid(self) -> str:
        """Get the guid currently used by the service.

        Returns:
            The guid string currently used by the service.
        """
        return self.__guid

    def get_apikey(self) -> str:
        """Get the apikey currently used by the service.

        Returns:
            The apikey string currently used by the service.
        """
        return self.__apikey

    def set_context(self, collection_id: str, environment_id: str, configuration_file: Optional[str] = None,
                    live_config_update_enabled: Optional[bool] = True, options: Optional[dict] = None):

        """Set the collection and environment value of the service.
        Args:

            collection_id (str): Id of the collection created in App Configuration service instance.
            environment_id (str): Id of the environment created in App Configuration service instance.
            configuration_file (str): [DEPRECATED]. Use the 'bootstrap_file' parameter in the options dict. Path to the JSON file which contains configuration details.
            live_config_update_enabled (bool): [DEPRECATED]. Use this kwarg as part of `options` dict. Set this value to false if the new configuration values shouldn't be fetched from the server. Make sure to provide a proper JSON file in the configuration_file path. By default, this value is enabled.
            options (dict): Dictionary object containing the optional arguments. See below for their usage.

        Optional arguments:
            persistent_cache_dir (str): Absolute path to a directory having write permission. The SDK will create a file - 'appconfiguration.json' file in the specified directory and it will be used as the persistent cache to store the App Configuration service information.
            bootstrap_file (str): Absolute path of configuration file. This parameter when passed along with `live_config_update_enabled` value will drive the SDK to use the configurations present in this file to perform feature & property evaluations.
            live_config_update_enabled (bool): live configurations update from the server. Set this value to `false` if the new configuration values shouldn't be fetched from the server.
        """

        if not self.__is_initialized:
            Logger.error(config_messages.COLLECTION_INIT_ERROR)
            return

        if not Validators.validate_string(collection_id):
            Logger.error(config_messages.COLLECTION_ID_VALUE_ERROR)
            return

        if not Validators.validate_string(environment_id):
            Logger.error(config_messages.ENVIRONMENT_ID_VALUE_ERROR)
            return

        default_options = {
            'persistent_cache_dir': None,
            'bootstrap_file': None,
            'live_config_update_enabled': live_config_update_enabled
        }

        if configuration_file is not None:
            Logger.info(config_messages.DEPRECATION_WARNING_MESSAGE)
            default_options['bootstrap_file'] = configuration_file

        if options is not None:
            if Validators.validate_set_context_options(options):
                for key in default_options:
                    if key in options:
                        default_options[key] = options[key]
            else:
                Logger.error(config_messages.SET_CONTEXT_OPTIONAL_ARGUMENTS_ERROR)
                return

        if not default_options['live_config_update_enabled'] and \
                (default_options['bootstrap_file'] is None or not Validators.validate_string(
                    default_options['bootstrap_file'])):
            Logger.error(config_messages.BOOTSTRAP_FILE_NOT_FOUND_ERROR)
            return

        self.__is_initialized_configuration = True

        self.__configuration_handler_instance.set_context(collection_id, environment_id, default_options)
        self.__load_data_now()

    def fetch_configurations(self):
        """Fetch the latest configurations"""
        if self.__is_initialized and self.__is_initialized_configuration:
            self.__load_data_now()
        else:
            Logger.error(config_messages.COLLECTION_INIT_ERROR)

    def __setup_configuration_handler(self):
        self.__configuration_handler_instance = ConfigurationHandler.get_instance()
        self.__configuration_handler_instance.init(region=self.__region, guid=self.__guid, apikey=self.__apikey,
                                                   override_server_host=self.override_server_host)

    def __load_data_now(self):
        if self.__is_loading:
            return
        self.__is_loading = True
        self.__configuration_handler_instance.load_data()
        self.__is_loading = False

    def register_configuration_update_listener(self, listener):
        """Register a listener for the Configuration changes.

        Args:
            listener: A method for listening to the Configuration changes
        """
        if self.__is_initialized and self.__is_initialized_configuration:
            self.__configuration_handler_instance.register_configuration_update_listener(listener)
        else:
            Logger.error(config_messages.COLLECTION_INIT_ERROR)

    def get_feature(self, feature_id: str) -> Feature:
        """Get the Feature with give Feature Id

        Args:
            feature_id: The Feature ID value.
        Returns:
            Feature object with the given feature_id. If the Feature is not available \
            then expect `None`.
        """
        if self.__is_initialized and self.__is_initialized_configuration:
            return self.__configuration_handler_instance.get_feature(feature_id)
        Logger.error(config_messages.COLLECTION_INIT_ERROR)
        return None

    def get_features(self) -> Dict[str, Feature]:
        """Get the list of Feature objects

        Returns:
            List of Feature objects
        """
        if self.__is_initialized and self.__is_initialized_configuration:
            return self.__configuration_handler_instance.get_features()
        Logger.error(config_messages.COLLECTION_INIT_ERROR)
        return None

    def get_properties(self) -> Dict[str, Property]:
        """Get the list of Property objects

        Returns:
            List of Property objects
        """
        if self.__is_initialized and self.__is_initialized_configuration:
            return self.__configuration_handler_instance.get_properties()
        Logger.error(config_messages.COLLECTION_INIT_ERROR)
        return None

    def get_property(self, property_id: str) -> Property:
        """Get the Property with give Property Id

        Args:
            property_id: The Property ID value.
        Returns:
            Property object with the given property_id. If the Property is \
            not available then expect `None`.
        """
        if self.__is_initialized and self.__is_initialized_configuration:
            return self.__configuration_handler_instance.get_property(property_id)
        Logger.error(config_messages.COLLECTION_INIT_ERROR)
        return None



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides methods to perform the input validations.
"""
import yaml
from schema import Schema, Optional, SchemaError, And
from .logger import Logger


class Validators:
    """Validator class"""

    @classmethod
    def validate_string(cls, value: str) -> bool:
        """Validate the string

        Args:
            value: value to be checked
        """
        return bool(value and value.strip())

    @classmethod
    def validate_yaml_string(cls, value: str):
        """Validate the yaml string and returns parsed yaml

        Args:
            value: yaml string to be checked and parsed
        """
        try:
            parsed_values = list(yaml.safe_load_all(value))
            if len(parsed_values) == 1:
                return parsed_values[0]
            return parsed_values
        except Exception as err:
            Logger.error("Error while parsing yaml string")
            Logger.error(err)
            return None

    @classmethod
    def validate_set_context_options(cls, options) -> bool:
        """Validates the options

        Args:
            options: optional keys to be checked
        """
        try:
            Schema({
                Optional('persistent_cache_dir'): And(str, lambda s: len(s) > 0),
                Optional('bootstrap_file'): And(str, lambda s: len(s) > 0),
                Optional('live_config_update_enabled'): bool
            }).validate(options)
            return True
        except SchemaError:
            return False



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides methods that perform the logging operations.
"""

import enum
import logging
import sys


def setup_custom_logger(name: str):
    """Set the logger class for the library

    Args:
        name: Name of the Logger
    """
    formatter = logging.Formatter(fmt='%(asctime)s AppConfiguration %(levelname)-8s %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')
    screen_handler = logging.StreamHandler(stream=sys.stdout)
    screen_handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(screen_handler)
    return logger


class Logger:
    """Logger for the library"""
    __is_debug = False

    __AppLogger = setup_custom_logger("appconfig_logger")

    class LoggerColors:
        """Logger Colors"""
        INFO = '\033[99m'
        ERROR = '\033[91m'
        SUCCESS = '\033[92m'
        WARNING = '\033[93m'
        DEBUG = '\033[99m'
        END = '\033[0m'

    class LEVEL(enum.Enum):
        """Logger Levels"""
        SUCCESS = 'SUCCESS'
        ERROR = 'ERROR'
        WARN = 'WARNING'
        INFO = 'INFO'
        DEBUG = 'DEBUG'

    @classmethod
    def set_debug(cls, value: bool):
        """Enable or disable the debug mode in Logger"""
        cls.__is_debug = value

    @classmethod
    def is_debug(cls):
        """Check whether the logger is in debug mode"""
        return cls.__is_debug

    @classmethod
    def info(cls, message):
        """Log info message"""
        message = str(message)
        cls.__AppLogger.info(cls.LoggerColors.INFO + message + cls.LoggerColors.END)

    @classmethod
    def error(cls, message):
        """Log error message"""
        message = str(message)
        cls.__AppLogger.error(cls.LoggerColors.ERROR + message + cls.LoggerColors.END)

    @classmethod
    def warning(cls, message):
        """Log warning message"""
        message = str(message)
        if cls.__is_debug:
            cls.__AppLogger.warning(cls.LoggerColors.WARNING + message + cls.LoggerColors.END)

    @classmethod
    def success(cls, message):
        """Log success message"""
        message = str(message)
        if cls.__is_debug:
            cls.__AppLogger.info(cls.LoggerColors.SUCCESS + message + cls.LoggerColors.END)

    @classmethod
    def debug(cls, message):
        """Log debug message"""
        message = str(message)
        if cls.__is_debug:
            cls.__AppLogger.debug(cls.LoggerColors.DEBUG + message + cls.LoggerColors.END)



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import unittest
from ibm_appconfiguration.configurations.models import Rule


class MyTestCase(unittest.TestCase):
    sut = None

    def set_up_email(self, value, attribute, operator):
        values = [value]
        rules = {
            'values': values,
            'operator': operator,
            'attribute_name': attribute
        }
        self.sut = Rule(rules)

    def test_test_rules(self):
        self.set_up_email('in.ibm.com', 'email', "endsWith")
        self.assertEqual(self.sut.get_attributes(), 'email')
        self.assertEqual(self.sut.get_operator(), 'endsWith')
        self.assertEqual(len(self.sut.get_values()), 1)
        self.assertNotEqual(self.sut.get_values()[0], 'in.test.com')
        self.assertEqual(self.sut.get_values()[0], 'in.ibm.com')

    def testNoEntity(self):
        self.set_up_email('in.ibm.com', 'email', "endsWith")
        client_attributes = {
            'email1': 'tester@in.ibm.com'
        }
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

    def test_evaluation_ends_with_string(self):
        self.set_up_email('in.ibm.com', 'email', 'endsWith')
        client_attributes = {
            'email': 'tester@in.ibm.com'
        }
        self.assertTrue(self.sut.evaluate_rule(client_attributes))
        client_attributes = {
            'email': 'tester@in.ibm.error'
        }
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

    def test_evaluation_is(self):
        self.set_up_email("123", "creditValues", "is")
        client_attributes = {
            'creditValues': '123'
        }
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

        self.set_up_email("123", "creditValues", "is")
        client_attributes = {
            'creditValues': 123
        }
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

        client_attributes = {
            'creditValues': 123
        }
        self.set_up_email(123, "creditValues", "is")
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

        client_attributes = {
            'creditValues': False
        }
        self.set_up_email(False, "creditValues", "is")
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

        client_attributes = {
            'creditValues': False
        }
        self.set_up_email("123", "creditValues", "is")
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

        self.set_up_email(123, "creditValues", "is")
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

        client_attributes = {
            'creditValues': False
        }
        self.set_up_email(True, "creditValues", "is")
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

        self.set_up_email(False, "creditValues", "is")
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

    def test_evaluation_startswith(self):
        self.set_up_email('user1', 'email', 'startsWith')
        client_attributes = {
            'email': 'user1@gm.com'
        }
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

    def test_evaluation_contains(self):
        self.set_up_email('+91', 'mobile', 'contains')
        client_attributes = {
            'mobile': '+91xxxxxxxxxx'
        }
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

        client_attributes = {
            'mobile': '+01xxxxxxxxxx'
        }
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

    def evaluation_common(self, type, val1, val2):
        self.set_up_email(val1, 'balance', type)
        client_attributes = {
            'balance': val2
        }
        self.assertTrue(self.sut.evaluate_rule(client_attributes))

        self.set_up_email(val2, 'balance', type)

        client_attributes = {
            'balance': val1
        }
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

        client_attributes = {
            'balance': '{0}'.format(val1)
        }
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

        client_attributes = {
            'balance': True
        }
        self.assertFalse(self.sut.evaluate_rule(client_attributes))

    def test_evaluation_greater_than(self):
        self.evaluation_common('greaterThan', 100, 200)

    def test_evaluation_lesser_than(self):
        self.evaluation_common('lesserThan', 200, 100)

    def test_evaluation_greater_than_equals(self):
        self.evaluation_common('greaterThanEquals', 100, 200)

    def test_evaluation_lesser_than_equals(self):
        self.evaluation_common('lesserThanEquals', 200, 100)

if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the methods to facilitate the API requests to the App Configuration service.
"""
import json as json_import
from typing import Optional, Union
from ibm_cloud_sdk_core import BaseService, DetailedResponse, ApiException
from .url_builder import URLBuilder
from .logger import Logger
from ibm_appconfiguration.version import __version__
from ..common import config_constants


class APIManager(BaseService):
    """The API Manager for the library"""

    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if APIManager.__instance is None:
            return APIManager()
        return APIManager.__instance

    def __init__(self):
        super().__init__(service_url=URLBuilder.get_base_url(),
                         authenticator=URLBuilder.get_iam_authenticator())
        APIManager.__instance = self

    def prepare_api_request(self,
                            method: str,
                            url: str,
                            data: Optional[Union[str, dict]] = None) -> DetailedResponse:
        """ Prepare the API call request

        Args:
            method: Method for the request
            url: Url of the request
            data: data to be send.
        Returns:
            return the DetailedResponse.
        """

        headers = {'Content-Type': 'application/json',
                   'User-Agent': '{0}/{1}'.format(config_constants.SDK_NAME, __version__)}
        if data and isinstance(data, dict):
            data = self.__remove_null_values(data)
            data = json_import.dumps(data)

        try:
            request = self.prepare_request(method=method,
                                           url=url,
                                           headers=headers,
                                           data=data)
            response = self.send(request)
            return response
        except ApiException as api_exception:
            return DetailedResponse(response=api_exception.message,
                                    headers=None,
                                    status_code=api_exception.code)
        except Exception as exception:
            Logger.debug(f'Error in service API call {str(exception)}')
            return DetailedResponse(status_code=400)

    def __remove_null_values(self, dictionary: dict) -> dict:
        """Create a new dictionary without keys mapped to null values.

        Args:
            dictionary: The dictionary potentially containing keys mapped to values of None.

        Returns:
            A dict with no keys mapped to None.
        """
        if isinstance(dictionary, dict):
            return {k: v for (k, v) in dictionary.items() if v is not None}
        return dictionary



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tkinter as tk
from ibm_appconfiguration import AppConfiguration, Feature, ConfigurationType
import config


class Window(tk.Tk):
    def __init__(self):
        super().__init__()

        self.__hasData = False
        self.title("App Configuration sample")
        self.label_text = tk.StringVar()
        self.label_text.set('Get your Configuration values after data is loaded')

        self.label = tk.Label(self, textvar=self.label_text)
        self.label.pack(fill=tk.BOTH, expand=1, padx=100, pady=30)

        hello_button = tk.Button(self, text="Get String Feature", command=lambda: self.fetch_feature('featurestring'))
        hello_button.pack(side=tk.LEFT, padx=(10, 20), pady=(0, 20))

        numeric_button = tk.Button(self, text="Get Numeric Feature",
                                   command=lambda: self.fetch_feature('featurenumeric'))
        numeric_button.pack(side=tk.LEFT, padx=(10, 20), pady=(0, 20))

        bool_button = tk.Button(self, text="Get Boolean Feature", command=lambda: self.fetch_feature('featurebool'))
        bool_button.pack(side=tk.LEFT, padx=(10, 20), pady=(0, 20))

        property_button = tk.Button(self, text="Property", command=lambda: self.fetch_property('numericproperty'))
        property_button.pack(side=tk.RIGHT, padx=(10, 20), pady=(0, 20))

        self.initialize_app()

    def fetch_property(self, property_id: str):
        if self.__hasData:
            self.label.configure(background="red")
            app_config = AppConfiguration.get_instance()
            property_obj = app_config.get_property(property_id)

            try:
                if property_obj:

                    entity_attributes = {
                        'city': 'Bangalore',
                        'country': 'India'
                    }
                    if property_obj.get_property_data_type() == ConfigurationType.NUMERIC:
                        val = property_obj.get_current_value("pvQr45", entity_attributes)
                        self.label_text.set(F"Your configurations property value is {val}")
                        self.configure(background="yellow")
                else:
                    print("No configurations")
            except Exception as err:
                print(err)
        else:
            self.label_text.set("Not loaded")
            self.label.configure(background="grey")

    def fetch_feature(self, feature_id: str):
        if self.__hasData:
            self.label.configure(background="red")
            app_config = AppConfiguration.get_instance()
            feature = app_config.get_feature(feature_id)
            try:
                if feature:

                    entity_attributes = {
                        'city': 'Bangalore',
                        'country': 'India'
                    }
                    if feature.get_feature_data_type() == ConfigurationType.STRING:
                        val = feature.get_current_value("pvQr45", entity_attributes)
                        self.label_text.set(F"Your configurations value is {val}")
                        self.configure(background="yellow")
                    elif feature.get_feature_data_type() == ConfigurationType.BOOLEAN:
                        val = feature.get_current_value("pvQr45", entity_attributes)
                        self.label_text.set(F"Your configurations value is {val}")
                        self.configure(background="green")
                    elif feature.get_feature_data_type() == ConfigurationType.NUMERIC:
                        val = feature.get_current_value("pvQr45", entity_attributes)
                        self.label_text.set(F"Your configurations value is {val}")
                        self.configure(background="black")
                else:
                    print("No configurations")
            except Exception as err:
                print(err)

        else:
            self.label_text.set("Not loaded")
            self.label.configure(background="red")

    def response(self):
        self.__hasData = True
        self.label_text.set('Get your Feature value NOW')

    def initialize_app(self):
        appconfig_client = AppConfiguration.get_instance()
        AppConfiguration.enable_debug(True)
        appconfig_client.init(region=AppConfiguration.REGION_US_SOUTH,
                              guid=config.GUID,
                              apikey=config.APIKEY)
        appconfig_client.set_context(collection_id=config.COLLECTION, environment_id=config.ENVIRONMENT)
        appconfig_client.register_configuration_update_listener(self.response)
        self.response()


if __name__ == "__main__":
    window = Window()
    window.mainloop()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file defines the constants used by the SDK.
"""

DEFAULT_SEGMENT_ID = '$$null$$'
DEFAULT_ENTITY_ID = '$$null$$'
DEFAULT_HTTP_TYPE = 'https://'
DEFAULT_WSS_TYPE = "wss://"
DEFAULT_BASE_URL = ".apprapp.cloud.ibm.com"
IAM_TEST_URL = "https://iam.test.cloud.ibm.com/identity/token"
IAM_PROD_URL = "https://iam.cloud.ibm.com/identity/token"
SDK_NAME = "appconfiguration-python-sdk"



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from ibm_appconfiguration.configurations.internal.utils.socket import Socket


class MyTestCase(unittest.TestCase):

    expected_message = ''
    expected_error = ''
    expected_closed_state = ''
    expected_open_state = ''

    def callback(self, message=None, error_state=None, closed_state=None, open_state=None):
        self.expected_message = message
        self.expected_error = error_state
        self.expected_closed_state = closed_state
        self.expected_open_state = open_state

    def test_socket(self):
        self.__socket = Socket()
        self.__socket.setup(
            url="",
            headers=[],
            callback=self.callback
        )

        self.assertIsNotNone(self.__socket)

        self.__socket.on_message("Socket message")
        self.assertEqual(self.expected_message, "Socket message")

        self.__socket.on_error("Error message")
        self.assertEqual(self.expected_error, "Error message")

        self.__socket.on_open()
        self.assertEqual(self.expected_open_state, "Opened the web_socket")

        self.__socket.on_close()
        self.assertEqual(self.expected_closed_state, "Closed the web_socket")




if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines the model of a Property defined in App Configuration service.
"""

from typing import Any
from ..internal.utils.logger import Logger
from ..internal.utils.validators import Validators
from .configuration_type import ConfigurationType


class Property:
    """Property object"""
    def __init__(self, property_list=dict):
        """
        @type property_list: dict
        """
        self.__name = property_list.get('name', '')
        self.__property_id = property_list.get('property_id', '')
        self.__segment_rules = property_list.get('segment_rules', list())
        self.__property_data = property_list
        self.__type = ConfigurationType(property_list.get('type') if property_list.get('type') is not None else ConfigurationType.NUMERIC)
        self.__format = property_list.get('format', None)
        self.__value = property_list.get('value', object)

    def get_property_name(self) -> str:
        """Get the Property name

        Returns:
            Return the Property name
        """
        return self.__name

    def get_value(self) -> Any:
        """Get the Property value

        Returns:
            Return the Property value
        """
        if self.__format == "YAML":
            return Validators.validate_yaml_string(self.__value)
        return self.__value

    def get_property_id(self) -> str:
        """Get the Property Id

        Returns:
            Return the Property Id
        """
        return self.__property_id

    def get_property_data_type(self) -> ConfigurationType:
        """Get the Property data type

        Returns:
            Return the Property data type
        """
        return self.__type

    def get_property_data_format(self) -> str:
        """Get the Property format type

        Returns:
            Return the Property format type
        """
        if self.__type == ConfigurationType.STRING and self.__format is None:
            return 'TEXT'
        return self.__format

    def get_segment_rules(self) -> list:
        """Get the Property segment rules

        Returns:
            Return the list of Property segment rules
        """
        return self.__segment_rules

    def get_current_value(self, entity_id: str, entity_attributes: {}) -> Any:
        """Get the evaluated value of the Property

        Args:
            entity_id: Id of the Entity
            entity_attributes: Entity attributes object
        Returns:
            Return the evaluated Property value
        """

        if not entity_id or entity_id == "":
            Logger.error("A valid entity id should be passed for this method.")
            return None
        from ibm_appconfiguration.configurations.configuration_handler import ConfigurationHandler
        configuration_handler = ConfigurationHandler.get_instance()
        return configuration_handler.property_evaluation(property_obj=self, entity_id=entity_id,
                                                         entity_attributes=entity_attributes)



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Internal class to handle the configuration.
"""
import os
from typing import Dict, List, Any
from threading import Timer, Thread
from ibm_appconfiguration.configurations.internal.common import config_messages, config_constants
from .internal.utils.logger import Logger
from .internal.utils.validators import Validators
from .models import Feature
from .models import SegmentRules
from .models import Segment
from .models import Property
from .internal.utils.file_manager import FileManager
from .internal.utils.metering import Metering
from .internal.utils.socket import Socket
from .internal.utils.url_builder import URLBuilder
from .internal.utils.connectivity import Connectivity
from .internal.utils.api_manager import APIManager


class ConfigurationHandler:
    """Internal class to handle the configuration"""
    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if ConfigurationHandler.__instance is None:
            return ConfigurationHandler()
        return ConfigurationHandler.__instance

    def __init__(self):

        """ Virtually private constructor. """
        self.__retry_count = 3
        if ConfigurationHandler.__instance is not None:
            raise Exception("ConfigurationHandler " + config_messages.SINGLETON_EXCEPTION)
        self.__collection_id = ''
        self.__environment_id = ''
        self.__apikey = ''
        self.__guid = ''
        self.__region = ''
        self.__is_initialized = False
        self.__configuration_update_listener = None
        self.__feature_map = dict()
        self.__property_map = dict()
        self.__segment_map = dict()
        self.__live_config_update_enabled = True
        ConfigurationHandler.__instance = self
        self.__retry_count = 3
        self.__retry_interval = 600
        self.__bootstrap_file = None
        self.__persistent_cache_dir = None
        self.__persistent_data = None
        self.__on_socket_retry = False
        self.__override_server_host = None
        self.__socket = None
        self.__connectivity = None
        self.__is_network_connected = True
        self.__api_manager = None

    def init(self, region: str,
             guid: str,
             apikey: str,
             override_server_host=str):
        """ Initialize the configuration.

        Args:
            region: Region name where the service instance is created.
            guid: GUID of the App Configuration service. Get it from the service credentials section of the dashboard
            apikey: ApiKey of the App Configuration service. Get it from the service credentials section of the dashboard
            override_server_host: Non public urls for testing purpose.
        """

        self.__apikey = apikey
        self.__guid = guid
        self.__region = region
        self.__override_server_host = override_server_host

        self.__feature_map = dict()
        self.__property_map = dict()
        self.__segment_map = dict()

    def set_context(self, collection_id: str, environment_id: str, options: dict):
        """Set the context for the configuration

        Args:
            collection_id: Id of the collection created in App Configuration service instance.
            environment_id: Id of the environment created in App Configuration service instance.
            options: Optional parameters such as persistent_cache_dir, bootstrap_file, & live_config_update_enabled
            Make sure to provide a proper JSON file in the bootstrap_file path.
            By default, this value is enabled.
        """

        self.__collection_id = collection_id
        self.__environment_id = environment_id
        URLBuilder.init_with_collection_id(collection_id=collection_id,
                                           guid=self.__guid,
                                           region=self.__region,
                                           environment_id=environment_id,
                                           override_server_host=self.__override_server_host,
                                           apikey=self.__apikey)
        Metering.get_instance().set_metering_url(URLBuilder.get_metering_url())
        self.__api_manager = APIManager.get_instance()
        self.__live_config_update_enabled = options['live_config_update_enabled']
        self.__bootstrap_file = options['bootstrap_file']
        self.__persistent_cache_dir = options['persistent_cache_dir']
        self.__is_initialized = True
        self.__check_network()

    def load_data(self):
        """Load the configuration data"""
        if not self.__is_initialized:
            Logger.error(config_messages.CONFIGURATION_HANDLER_INIT_ERROR)
            return
        if self.__persistent_cache_dir:
            self.__persistent_data = FileManager.read_files(
                file_path=os.path.join(self.__persistent_cache_dir, 'appconfiguration.json'))
            if self.__persistent_data is not None:
                self.__load_configurations(self.__persistent_data)
            if not os.access(self.__persistent_cache_dir, os.W_OK):
                Logger.error(config_messages.ERROR_NO_WRITE_PERMISSION)
                return
        if self.__bootstrap_file:
            if self.__persistent_cache_dir:
                if self.__persistent_data is None or len(self.__persistent_data) == 0:
                    bootstrap_file_data = FileManager.read_files(file_path=self.__bootstrap_file)
                    if bootstrap_file_data is not None:
                        self.__load_configurations(bootstrap_file_data)
                    else:
                        Logger.error("Error reading bootstrap file data")
                        return
                    self.__write_to_persistent_storage(bootstrap_file_data, self.__persistent_cache_dir)
                    if self.__configuration_update_listener and callable(self.__configuration_update_listener):
                        self.__configuration_update_listener()
                else:
                    if self.__configuration_update_listener and callable(self.__configuration_update_listener):
                        self.__configuration_update_listener()
            else:
                bootstrap_file_data = FileManager.read_files(file_path=self.__bootstrap_file)
                if bootstrap_file_data is not None:
                    self.__load_configurations(bootstrap_file_data)
                else:
                    Logger.error("Error reading bootstrap file data")
                    return
        if self.__live_config_update_enabled:
            self.__fetch_config_data()
        else:
            if self.__socket:
                self.__socket.cancel()

    def register_configuration_update_listener(self, listener):
        """Register the listener

        Args:
            listener: Listener for the configuration update.
        """
        if callable(listener):
            if self.__is_initialized:
                self.__configuration_update_listener = listener
            else:
                Logger.error(config_messages.CONFIGURATION_HANDLER_INIT_ERROR)
        else:
            Logger.error(config_messages.CONFIGURATION_HANDLER_METHOD_ERROR)

    def __check_network(self):
        if self.__live_config_update_enabled:
            if self.__connectivity is None:
                self.__connectivity = Connectivity.get_instance()
                self.__connectivity.add_connectivity_listener(self.__network_listener)
                self.__connectivity.check_connection()
        else:
            self.__connectivity = None

    def __network_listener(self, is_connected: bool):
        if not self.__live_config_update_enabled:
            self.__connectivity = None
            return

        if is_connected:
            if not self.__is_network_connected:
                self.__is_network_connected = True
                self.__fetch_config_data()
        else:
            Logger.debug(config_messages.NO_INTERNET_CONNECTION_ERROR)
            self.__is_network_connected = False

    def get_properties(self) -> Dict[str, Property]:
        """Get the list of Property objects

        Returns:
            List of Property objects
        """
        return self.__property_map

    def get_property(self, property_id: str) -> Property:
        """Get the Property with give Property Id

        Args:
            property_id: The Property ID value.
        Returns:
            Property object with the given property_id. If the Property is \
            not available then expect `None`.
        """
        if property_id in self.__property_map:
            return self.__property_map.get(property_id)
        Logger.error(config_messages.PROPERTY_INVALID + property_id)
        return None

    def get_features(self) -> Dict[str, Feature]:
        """Get the list of Feature objects

        Returns:
            List of Feature objects
        """
        return self.__feature_map

    def get_feature(self, feature_id: str) -> Feature:
        """Get the Feature with give Feature Id

        Args:
            feature_id: The Feature ID value.
        Returns:
            Feature object with the given feature_id. If the Feature is not available \
            then expect `None`.
        """
        if feature_id in self.__feature_map:
            return self.__feature_map.get(feature_id)
        Logger.error(config_messages.FEATURE_INVALID + feature_id)
        return None

    def __fetch_config_data(self):
        if self.__is_initialized:
            self.__fetch_from_api()
            self.__on_socket_retry = False
            config_thread = Thread(target=self.__start_web_socket, args=())
            config_thread.daemon = True
            config_thread.start()

    def __start_web_socket(self):
        bearer_token = URLBuilder.get_iam_authenticator().token_manager.get_token()
        headers = {
            'Authorization': 'Bearer ' + bearer_token
        }
        if self.__socket:
            self.__socket.cancel()
            self.__socket = None
        self.__socket = Socket()
        self.__socket.setup(
            url=URLBuilder.get_web_socket_url(),
            headers=headers,
            callback=self.__on_web_socket_callback
        )

    def __load_configurations(self, data: dict):
        if len(data) != 0:
            if 'features' in data:
                self.__feature_map = dict()
                try:
                    all_feature_list: List = data.get('features')
                    for i, feature in enumerate(all_feature_list):
                        feature_obj = Feature(feature)
                        self.__feature_map[feature_obj.get_feature_id()] = feature_obj
                except Exception as err:
                    Logger.debug(err)

            if 'properties' in data:
                self.__property_map = dict()
                try:
                    all_property_list: List = data.get('properties')
                    for i, property_list in enumerate(all_property_list):
                        property_obj = Property(property_list)
                        self.__property_map[property_obj.get_property_id()] = property_obj
                except Exception as err:
                    Logger.debug(err)

            if 'segments' in data:
                self.__segment_map = dict()
                try:
                    segment_list: List = data.get('segments')
                    for i, segment in enumerate(segment_list):
                        segment: dict = segment_list[i]
                        segment_obj = Segment(segment)
                        self.__segment_map[segment_obj.get_segment_id()] = segment_obj
                except Exception as err:
                    Logger.debug(err)

    def record_valuation(self, property_id, feature_id, entity_id, evaluated_segment_id):
        """Record the evaluation data.

        Args:
            property_id: Id of the Property
            feature_id: Id of the Feature
            entity_id: Id of the Entity
            evaluated_segment_id: Id of the Segment
        """
        Metering.get_instance().add_metering(
            guid=self.__guid,
            environment_id=self.__environment_id,
            collection_id=self.__collection_id,
            entity_id=entity_id,
            segment_id=evaluated_segment_id,
            feature_id=feature_id,
            property_id=property_id
        )

    def property_evaluation(self, property_obj: Property, entity_id: str,
                            entity_attributes: dict = None) -> Any:
        """Property evaluation method

        Args:
            property_obj: Property object
            entity_id: Entity Id
            entity_attributes: Entity attributes object
        Returns:
            Return evaluated value
        """

        result_dict = {
            'evaluated_segment_id': config_constants.DEFAULT_SEGMENT_ID,
            'value': None
        }

        try:
            if entity_attributes is None or len(entity_attributes) <= 0:
                return property_obj.get_value()

            segment_rules = property_obj.get_segment_rules()
            if len(segment_rules) > 0:
                rules_map = self.__parse_rules(segment_rules)
                result_dict = self.__evaluate_rules(rules_map, entity_attributes,
                                                    property_obj=property_obj)
                # if segment is null or segment value is default then yaml is auto converted
                if property_obj.get_property_data_format() == "YAML" and type(result_dict['value']) == str:
                    return Validators.validate_yaml_string(result_dict['value'])
                return result_dict['value']
            return property_obj.get_value()

        finally:
            property_id = property_obj.get_property_id()
            self.record_valuation(property_id=property_id, feature_id=None, entity_id=entity_id,
                                  evaluated_segment_id=result_dict['evaluated_segment_id'])

    def feature_evaluation(self, feature: Feature, is_enabled: bool, entity_id: str,
                           entity_attributes: dict = None) -> Any:
        """Feature evaluation method

        Args:
            feature: Feature object
            is_enabled: Feature object's "enabled" value (True/False)
            entity_id: Entity Id
            entity_attributes: Entity attributes object
        Returns:
            Return evaluated value
        """
        result_dict = {
            'evaluated_segment_id': config_constants.DEFAULT_SEGMENT_ID,
            'value': None
        }
        try:
            if is_enabled:

                if entity_attributes is None or len(entity_attributes) <= 0:
                    return feature.get_enabled_value()

                segment_rules = feature.get_segment_rules()
                if len(segment_rules) > 0:
                    rules_map = self.__parse_rules(segment_rules)
                    result_dict = self.__evaluate_rules(rules_map, entity_attributes, feature=feature)
                    # if segment is null or segment value is default then yaml is auto converted
                    if feature.get_feature_data_format() == "YAML" and type(result_dict['value']) == str:
                        return Validators.validate_yaml_string(result_dict['value'])
                    return result_dict['value']
                return feature.get_enabled_value()
            return feature.get_disabled_value()
        finally:
            feature_id = None if feature is None else feature.get_feature_id()
            self.record_valuation(property_id=None, feature_id=feature_id, entity_id=entity_id,
                                  evaluated_segment_id=result_dict['evaluated_segment_id'])

    def __evaluate_rules(self, rules_map: dict,
                         entity_attributes: {},
                         feature: Feature = None,
                         property_obj: Property = None) -> dict:
        result_dict = {
            'evaluated_segment_id': config_constants.DEFAULT_SEGMENT_ID,
            'value': None
        }
        for i in range(1, len(rules_map) + 1):
            segment_rule = rules_map[i]
            if segment_rule is not None:
                for level in range(0, len(segment_rule.get_rules())):
                    try:
                        rule: dict = segment_rule.get_rules()[level]
                        segments: List = rule.get('segments')
                        for _, segment_key in enumerate(segments):
                            if self.__evaluate_segment(segment_key, entity_attributes):
                                result_dict['evaluated_segment_id'] = segment_key
                                if segment_rule.get_value() == "$default":
                                    result_dict[
                                        'value'] = feature.get_enabled_value() if feature is not \
                                                                                  None else property_obj.get_value()
                                else:
                                    result_dict['value'] = segment_rule.get_value()
                                return result_dict
                    except Exception as err:
                        Logger.debug(err)

        result_dict['value'] = feature.get_enabled_value() if feature is not None else property_obj.get_value()
        return result_dict

    def __evaluate_segment(self, segment_key: str, entity_attributes: dict) -> bool:
        if segment_key in self.__segment_map:
            segment: Segment = self.__segment_map[segment_key]
            return segment.evaluate_rule(entity_attributes)
        return False

    def __parse_rules(self, segment_rules: List) -> dict:
        rule_map = dict()
        for _, rules in enumerate(segment_rules):
            try:
                rules_obj = SegmentRules(rules)
                rule_map[rules_obj.get_order()] = rules_obj
            except Exception as err:
                Logger.debug(err)
        return rule_map

    def __write_to_persistent_storage(self, json: dict, file_path: str):
        FileManager.store_files(json, os.path.join(file_path, 'appconfiguration.json'))

    def __fetch_from_api(self):
        if self.__is_initialized:
            self.__retry_count -= 1

            response = self.__api_manager.prepare_api_request(method="GET", url=URLBuilder.get_config_url())
            status_code = response.get_status_code()

            if 200 <= status_code <= 299:
                Logger.info(config_messages.CONFIGURATIONS_FETCH_SUCCESS)
                response_data = response.get_result()
                try:
                    response_data = dict(response_data)
                    self.__load_configurations(response_data)  # load response to cache maps
                    if self.__configuration_update_listener and callable(self.__configuration_update_listener):
                        self.__configuration_update_listener()
                    if self.__persistent_cache_dir:
                        file_write_thread = Thread(target=self.__write_to_persistent_storage,
                                                   args=(response_data, self.__persistent_cache_dir,))
                        file_write_thread.daemon = True
                        file_write_thread.start()
                except Exception as exception:
                    Logger.error(f'error while while fetching {exception}')
            else:
                Logger.error(response.get_result())
                if self.__retry_count > 0:
                    self.__fetch_from_api()
                else:
                    self.__retry_count = 3
                    Logger.info(config_messages.RETRY_AFTER_TEN_MINUTES)
                    timer = Timer(self.__retry_interval, self.__fetch_from_api)
                    timer.daemon = True
                    timer.start()
        else:
            Logger.debug(config_messages.CONFIGURATION_HANDLER_INIT_ERROR)

    def __on_web_socket_callback(self, message=None, error_state=None,
                                 closed_state=None, open_state=None):
        if message:
            self.__fetch_from_api()
            Logger.debug(f'Received message from socket. {message}')
        elif error_state:
            Logger.error(f'Received error from socket. {error_state}')
            self.__start_web_socket()
        elif closed_state:
            Logger.error('Received close connection from socket.')
            if self.__socket is not None:
                self.__on_socket_retry = True
                self.__start_web_socket()
        elif open_state:
            if self.__on_socket_retry:
                self.__on_socket_retry = False
                self.__fetch_from_api()
            Logger.debug('Received opened connection from socket.')
        else:
            Logger.error('Unknown Error inside the socket connection.')



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import unittest
import os
from ibm_appconfiguration import AppConfiguration
from ibm_appconfiguration.configurations.internal.utils.url_builder import URLBuilder


class MyTestCase(unittest.TestCase):

    def setUp(self):
        URLBuilder.set_auth_type(False)

    def test_configuration(self):
        sut1 = AppConfiguration.get_instance()
        sut2 = AppConfiguration.get_instance()
        self.assertEqual(sut1, sut2)

    def test_configuration_values(self):
        sut1 = AppConfiguration.get_instance()

        sut1.init(None, "guid_value", "apikey_value")
        self.assertIsNotNone(sut1.get_region())

        sut1.init('region', None, "apikey_value")
        self.assertIsNotNone(sut1.get_guid())

        sut1.init('region', "guid_value", None)
        self.assertIsNotNone(sut1.get_apikey())

        sut1.init('region', "guid_value", "apikey_value")
        self.assertEqual(sut1.get_guid(), "guid_value")
        self.assertEqual(sut1.get_apikey(), "apikey_value")
        self.assertEqual(sut1.get_region(), "region")

    def test_configuration_fetch(self):
        sut1 = AppConfiguration.get_instance()
        sut1.set_context("", "")
        self.assertIsNotNone(sut1.get_apikey())

    def test_configuration_fetch_feature_data(self):
        sut1 = AppConfiguration.get_instance()
        sut1.fetch_configurations()

    def response(self):
        print('Get your Feature value NOW')

    def test_configuration_register_features_update_listener(self):
        sut1 = AppConfiguration.get_instance()
        sut1.register_configuration_update_listener(self.response)

    def test_configuration_get_feature(self):
        sut1 = AppConfiguration.get_instance()
        self.assertIsNone(sut1.get_feature("FeatureId"))

    def test_configuration_get_features(self):
        sut1 = AppConfiguration.get_instance()
        sut1.init('region', "guid_value", "apikey_value")
        sut1.enable_debug(True)

        FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'user.json')
        sut1.set_context("collectionId", "environmentId", configuration_file=FILE, live_config_update_enabled=False)
        time.sleep(2.5)

        self.assertEqual(len(sut1.get_features()), 3)

    def test_configuration_get_features_Dict(self):
        sut1 = AppConfiguration.get_instance()
        self.assertIsNotNone(sut1.get_features())


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines the model of a feature flag defined in App Configuration service.
"""

from typing import Any
from ..internal.common import config_constants
from ..internal.utils.logger import Logger
from ..internal.utils.validators import Validators
from .configuration_type import ConfigurationType


class Feature:
    """Feature object"""

    def __init__(self, feature_list=dict):
        """
        @type feature_list: dict
        """
        self.__enabled = feature_list.get('enabled', False)
        self.__name = feature_list.get('name', '')
        self.__feature_id = feature_list.get('feature_id', '')
        self.__segment_rules = feature_list.get('segment_rules', list())
        self.__feature_data = feature_list
        self.__type = ConfigurationType(
            feature_list.get('type') if feature_list.get('type') is not None else ConfigurationType.NUMERIC)
        self.__format = feature_list.get('format', None)
        self.__disabled_value = feature_list.get('disabled_value', object)
        self.__enabled_value = feature_list.get('enabled_value', object)

    def get_feature_name(self) -> str:
        """Get the Feature name"""
        return self.__name

    def get_disabled_value(self) -> Any:
        """Get the Feature disabled value"""
        if self.__format == "YAML":
            return Validators.validate_yaml_string(self.__disabled_value)
        return self.__disabled_value

    def get_enabled_value(self) -> Any:
        """Get the Feature enabled value"""
        if self.__format == "YAML":
            return Validators.validate_yaml_string(self.__enabled_value)
        return self.__enabled_value

    def get_feature_id(self) -> str:
        """Get the Feature Id"""
        return self.__feature_id

    def get_feature_data_type(self) -> ConfigurationType:
        """Get the Feature data type"""
        return self.__type

    def get_feature_data_format(self) -> str:
        """Get the Feature data format"""
        if self.__type == ConfigurationType.STRING and self.__format is None:
            return 'TEXT'
        return self.__format

    def is_enabled(self) -> bool:
        """Check the Feature is enabled or not"""
        from ibm_appconfiguration.configurations.configuration_handler import ConfigurationHandler
        feature_handler = ConfigurationHandler.get_instance()
        feature_handler.record_valuation(property_id=None, feature_id=self.__feature_id,
                                         entity_id=config_constants.DEFAULT_ENTITY_ID,
                                         evaluated_segment_id=config_constants.DEFAULT_SEGMENT_ID)
        return self.__enabled

    def get_segment_rules(self) -> list:
        """Get the Feature segment_rules"""
        return self.__segment_rules

    def get_current_value(self, entity_id: str, entity_attributes: {}) -> Any:
        """Get the Feature evaluated value

        Args:
            entity_id: Id of the Entity
            entity_attributes: Entity attributes object
        Returns:
            Return the evaluated Feature value
        """
        if not entity_id or entity_id == "":
            Logger.error("A valid entity id should be passed for this method.")
            return None
        from ibm_appconfiguration.configurations.configuration_handler import ConfigurationHandler
        feature_handler = ConfigurationHandler.get_instance()
        return feature_handler.feature_evaluation(feature=self, is_enabled=self.__enabled, entity_id=entity_id,
                                                  entity_attributes=entity_attributes)



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from ibm_appconfiguration.configurations.internal.utils.metering import Metering


class MyTestCase(unittest.TestCase):

    def test_metering(self):

        metering = Metering.get_instance()
        metering.set_repeat_calls(False)
        # test features
        metering.add_metering(guid="guid1", environment_id="environment_id1", collection_id="collection_id1",
                              entity_id="id_1", segment_id="segment_id1", feature_id='feature_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id1", collection_id="collection_id2",
                              entity_id="id_1", segment_id="segment_id1", feature_id='feature_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id1", collection_id="collection_id2",
                              entity_id="id_1", segment_id="segment_id1", feature_id='feature_id1')

        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id1",
                              entity_id="id_1", segment_id="segment_id1", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_1", segment_id="segment_id1", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_1", segment_id="$$null$$", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_6", segment_id="$$null$$", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_3", segment_id="$$null$$", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_4", segment_id="$$null$$", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_1", segment_id="$$null$$", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_1", segment_id="$$null$$", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_2", segment_id="$$null$$", property_id='property_id1')
        metering.add_metering(guid="guid1", environment_id="environment_id2", collection_id="collection_id2",
                              entity_id="id_5", segment_id="$$null$$", property_id='property_id1')

        result = metering.send_metering()
        print("result")

        # print(result)
        self.assertEqual(len(result["guid1"]), 4)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os
from ibm_appconfiguration.configurations.internal.utils.file_manager import FileManager


class MyTestCase(unittest.TestCase):
    file_path = ''

    def setUp(self) -> None:
        self.file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'list.json')

    def test_file_store(self):
        data = {
            "features": [
                {
                    "name": "defaultFeature123",
                    "feature_id": "defaultfeature",
                    "type": "BOOLEAN",
                    "enabled_value": True,
                    "disabled_value": False,
                    "segment_rules": [
                        {
                            "rules": [
                                {
                                    "segments": [
                                        "kg92d3wa"
                                    ]
                                }
                            ],
                            "value": False,
                            "order": 1
                        }
                    ],
                    "segment_exists": True,
                    "isEnabled": True
                }
            ],
            "collection": {
                "name": "appCrash",
                "collection_id": "appcrash"
            },
            "segments": [
                {
                    "name": "defaultSeg",
                    "segment_id": "kg92d3wa",
                    "rules": [
                        {
                            "values": [
                                "dev"
                            ],
                            "operator": "contains",
                            "attribute_name": "email"
                        }
                    ]
                }
            ]
        }
        self.assertTrue(FileManager.store_files(data, self.file_path))

        expected_data = FileManager.read_files(self.file_path)

        self.assertIsNotNone(expected_data)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from http.server import HTTPServer, BaseHTTPRequestHandler
from ibm_appconfiguration import AppConfiguration, Feature, ConfigurationType
import config
import time
import secrets

has_data = True


def response():
    print('Get your Feature value NOW')


def setup() -> None:
    appconfig_client = AppConfiguration.get_instance()
    appconfig_client.init(region=AppConfiguration.REGION_US_SOUTH,
                          guid=config.GUID,
                          apikey=config.APIKEY)
    appconfig_client.set_context(collection_id=config.COLLECTION, environment_id=config.ENVIRONMENT)


def fetch_feature(feature_id: str) -> str:
    if has_data:
        app_config = AppConfiguration.get_instance()
        feature = app_config.get_feature(feature_id)
        try:
            if feature:
                entity_attributes = {
                    'city': 'Bangalore',
                    'country': 'India'
                }
                if feature.get_feature_data_type() == ConfigurationType.STRING:
                    val = feature.get_current_value(id='pvQr45', entity_attributes=entity_attributes)
                    return f"Your feature value is {val}"
                elif feature.get_feature_data_type() == ConfigurationType.BOOLEAN:
                    val = feature.get_current_value('pvQr45', entity_attributes)
                    return f"Your feature value is {val}"
                elif feature.get_feature_data_type() == ConfigurationType.NUMERIC:
                    val = feature.get_current_value('pvQr45', entity_attributes)
                    return f"Your feature value is {val}"
            else:
                return "Not loaded"
        except Exception as err:
            print(err)
            return "Not loaded"

    else:
        return "Not loaded"


def _html(message):
    """This just generates an HTML document that includes `message`
    in the body. Override, or re-write this do do more interesting stuff.
    """
    content = f"<html><body><h1>{message}</h1></body></html>"
    return content.encode("utf8")  # NOTE: must return a bytes object!


class MockServer(BaseHTTPRequestHandler):

    def _set_headers(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_GET(self):

        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()

        locations = ['Texas', 'California']
        app_config = AppConfiguration.get_instance()
        location = secrets.choice(locations)

        feature = app_config.get_feature("featurestring")

        entity_attributes = {
            'id': 'pvqr',
            "location": location
        }
        value = feature.get_current_value("pvQr45", entity_attributes)

        is_valid = True
        if value == "Hello Googler":
            is_valid = "Texas" == location
        elif value == "Hello IBMer":
            is_valid = "California" == location

        self.wfile.write(json.dumps({'is_valid': is_valid, 'featurevalue': value,
                                     'time': round(time.time() * 1000)}).encode('utf-8'))

    def do_HEAD(self):
        self._set_headers()

    def do_POST(self):
        value = "POST!"
        if has_data:
            value = fetch_feature("featurestring")
        self._set_headers()
        self.wfile.write(_html(value))


def run(server_class=HTTPServer, handler_class=MockServer, addr="localhost", port=8000):
    server_address = (addr, port)
    httpd = server_class(server_address, handler_class)
    print(f"Starting httpd server on {addr}:{port}")
    httpd.serve_forever()


if __name__ == '__main__':
    setup()
    run()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from ibm_appconfiguration.configurations.internal.utils.metering import Metering
from ibm_appconfiguration.configurations.models import Feature, ConfigurationType


class MyTestCase(unittest.TestCase):
    sut = None

    def test_empty_object(self):
        sut = Feature(dict())
        self.assertFalse(sut.is_enabled())
        self.assertEqual(sut.get_feature_name(), "")
        self.assertEqual(sut.get_feature_id(), "")
        self.assertEqual(len(sut.get_segment_rules()), 0)
        self.assertEqual(sut.get_feature_data_type(), ConfigurationType.NUMERIC)
        self.assertIsNotNone(sut.get_disabled_value())
        self.assertIsNotNone(sut.get_enabled_value())

    def test_feature(self):
        feature = {
            "name": "testF5",
            "feature_id": "testf5",
            "description": "Feature flags to enable testing",
            "type": "BOOLEAN",
            "enabled": True,
            "enabled_value": True,
            "disabled_value": False,
            "tags": "version: 1.1, pre-release",
            "segment_rules": [{
                "rules": [
                    {
                        "segments": [
                            "300"
                        ]
                    }
                ],
                "value": "$default",
                "order": 1
            }
            ]
        }

        sut = Feature(feature)
        self.assertTrue(sut.is_enabled())
        self.assertEqual(sut.get_feature_name(), "testF5")
        self.assertEqual(sut.get_feature_id(), "testf5")
        Metering.get_instance().set_repeat_calls(False)

        value = sut.get_current_value("id1", {"email": "test.dev@tester.com"})
        self.assertEqual(value, True)

    # enabled and disabled value contains valid yaml
    def test_feature_yaml1(self):
        feature = {
            "name": "testF5",
            "feature_id": "testf5",
            "description": "Feature flags to enable testing",
            "type": "STRING",
            "format": "YAML",
            "enabled": True,
            "enabled_value": "key1: enabled_value",
            "disabled_value": "key2: disabled_value\n---\nkey22: disabled_value22",
            "tags": "version: 1.1, pre-release",
            "segment_rules": [{
                "rules": [
                    {
                        "segments": [
                            "300"
                        ]
                    }
                ],
                "value": "$default",
                "order": 1
            }
            ]
        }

        sut = Feature(feature)
        self.assertTrue(sut.is_enabled())
        self.assertEqual(sut.get_feature_name(), "testF5")
        self.assertEqual(sut.get_feature_id(), "testf5")
        self.assertEqual(sut.get_feature_data_format(), "YAML")
        self.assertEqual(sut.get_enabled_value()['key1'], "enabled_value")
        self.assertEqual(sut.get_disabled_value()[1]['key22'], "disabled_value22")
        Metering.get_instance().set_repeat_calls(False)

        value = sut.get_current_value("id1", {"email": "test.dev@tester.com"})
        self.assertEqual(value['key1'], "enabled_value")

    # enabled and disabled value contains invalid yaml documents and invalid entity id is passed
    def test_feature_yaml2(self):
        feature = {
            "name": "testF5",
            "feature_id": "testf5",
            "description": "Feature flags to enable testing",
            "type": "STRING",
            "format": "YAML",
            "enabled": True,
            "enabled_value": "key1: enabled_value1\n--\n",
            "disabled_value": "key2: disabled_value2\n--\n",
            "tags": "version: 1.1, pre-release"
        }

        sut = Feature(feature)
        self.assertTrue(sut.is_enabled())
        self.assertEqual(sut.get_feature_name(), "testF5")
        self.assertEqual(sut.get_feature_id(), "testf5")
        self.assertEqual(sut.get_feature_data_format(), "YAML")
        self.assertEqual(sut.get_enabled_value(), None)
        self.assertEqual(sut.get_disabled_value(), None)
        Metering.get_instance().set_repeat_calls(False)

        value = sut.get_current_value("", {"email": "test.dev@tester.com"})
        self.assertEqual(value, None)

    # enabled and disabled value contains valid json
    def test_feature_json(self):
        feature = {
            "name": "testF5",
            "feature_id": "testf5",
            "description": "Feature flags to enable testing",
            "type": "STRING",
            "format": "JSON",
            "enabled": True,
            "enabled_value": {
                "key1": "value is enabled"
            },
            "disabled_value": {
                "key2": "value is disabled"
            },
            "tags": "version: 1.1, pre-release"
        }

        sut = Feature(feature)
        self.assertTrue(sut.is_enabled())
        self.assertEqual(sut.get_feature_name(), "testF5")
        self.assertEqual(sut.get_feature_id(), "testf5")
        self.assertEqual(sut.get_feature_data_format(), "JSON")
        self.assertEqual(sut.get_enabled_value()['key1'], "value is enabled")
        self.assertEqual(sut.get_disabled_value()['key2'], "value is disabled")
        Metering.get_instance().set_repeat_calls(False)



if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from ibm_appconfiguration.configurations.internal.utils.url_builder import URLBuilder


class MyTestCase(unittest.TestCase):

    def test_url_builder(self):
        URLBuilder.init_with_collection_id(collection_id="collection_id",
                                           guid="guid",
                                           environment_id="environment_id",
                                           region="region",
                                           override_server_host="",
                                           apikey="")
        expected_config_url = '/feature/v1/instances/guid/collections/collection_id/config?environment_id=environment_id'
        expected_socket_url = 'wss://region.apprapp.cloud.ibm.com/apprapp/wsfeature?instance_id=guid&collection_id=collection_id&environment_id=environment_id'
        expected_metering_url = '/events/v1/instances/'

        self.assertEqual(URLBuilder.get_config_url(), expected_config_url)
        self.assertEqual(URLBuilder.get_web_socket_url(), expected_socket_url)
        self.assertEqual(URLBuilder.get_metering_url(), expected_metering_url)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides methods that perform the store and retrieve operations on the
file based cache of the SDK.
"""

import fcntl
import json
from typing import Optional, Any

from .logger import Logger


class FileManager:
    """FileManager to handle the cache"""

    @classmethod
    def store_files(cls, json_data: {}, file_path: str) -> bool:
        """Store the file

        Args:
            json_data: Data to be stored.
            file_path: File path for the cache.
        """
        try:
            with open(file_path, 'w') as cache:
                fcntl.flock(cache, fcntl.LOCK_EX | fcntl.LOCK_NB)
                json.dump(json_data, cache)
                fcntl.flock(cache, fcntl.LOCK_UN)
                return True
        except Exception as err:
            Logger.error(err)
            return False

    @classmethod
    def read_files(cls, file_path: str) -> Optional[Any]:
        """
        Read the data from the given path.

        Args:
            file_path: Path of the file
        Returns:
            Dictionary.
        """

        try:
            with open(file_path, 'r') as file:
                fcntl.flock(file, fcntl.LOCK_EX | fcntl.LOCK_NB)
                data = json.load(file)
                fcntl.flock(file, fcntl.LOCK_UN)
                return data
        except Exception as err:
            Logger.error(err)
            return None



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides methods to perform operations on the websocket connection to the server.
"""
import ssl
import websocket


class Socket:
    """Class to handle the Web socket"""
    def __init__(self):
        self.__callback = None
        self.ws_client = None

    def setup(self, url, headers, callback):
        """ Setup the socket.

        Args:
            url: Url for the socket
            headers: Headers for the socket.
            callback: Callback for the socket.
        """
        self.__callback = callback
        self.ws_client = websocket.WebSocketApp(
            url,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
            header=headers
        )
        self.ws_client.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})

    def on_message(self, message):
        """Socket on-message

        Args:
            message: Message object from the socket
        """
        if message == 'test message':
            return
        self.__callback(message=message)

    def on_error(self, error):
        """Socket on-error

        Args:
            error: Error object from the socket
        """
        self.__callback(error_state=error)
        self.ws_client.close()

    def on_close(self):
        """Socket on-close call"""
        self.__callback(closed_state='Closed the web_socket')

    def on_open(self):
        """Socket on-open call"""
        self.__callback(open_state='Opened the web_socket')

    def cancel(self):
        """
        Socket cancel.
        """
        self.ws_client.close()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import unittest
from ibm_appconfiguration.configurations.models import SegmentRules


class TestSegmentRules(unittest.TestCase):

    def test_empty_object(self):
        sut = SegmentRules(dict())
        self.assertEqual(sut.get_order(), 1)
        self.assertEqual(len(sut.get_rules()), 0)
        self.assertIsNotNone(sut.get_value())

    def test_object(self):
        segments = ["kg92d3wa"]
        json_seg_obj = {"segments": segments}
        rules = [json_seg_obj]
        segment_rules = {
            "rules": rules,
            "value": "IBM user",
            "order": 1
        }

        sut = SegmentRules(segment_rules)
        self.assertEqual(sut.get_order(), 1)
        self.assertEqual(len(sut.get_rules()), 1)
        self.assertEqual(sut.get_value(), "IBM user")


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from ibm_appconfiguration.configurations.models import Property, ConfigurationType


class MyTestCase(unittest.TestCase):
    sut = None

    def test_empty_property(self):
        sut = Property(dict())
        self.assertEqual(sut.get_property_name(), "")
        self.assertEqual(sut.get_property_id(), "")
        self.assertEqual(len(sut.get_segment_rules()), 0)
        self.assertEqual(sut.get_property_data_type(), ConfigurationType.NUMERIC)
        self.assertIsNotNone(sut.get_value())

    def test_property(self):
        property = {
            "name": "numericProperty",
            "property_id": "numericproperty",
            "description": "testing prop",
            "value": 10,
            "type": "NUMERIC",
            "tags": "test",
            "segment_rules": [
                {
                    "rules": [
                        {
                            "segments": [
                                "keuyclvf"
                            ]
                        }
                    ],
                    "value": 81,
                    "order": 1
                }
            ],
            "collections": [{
                "collection_id": "appcrash"
            }]
        }

        sut = Property(property)
        self.assertEqual(sut.get_property_name(), "numericProperty")
        self.assertEqual(sut.get_property_id(), "numericproperty")
        self.assertEqual(sut.get_value(), 10)
        self.assertEqual(sut.get_property_data_type(), ConfigurationType.NUMERIC)
        value = sut.get_current_value("id1", {"email": "test.dev@tester.com"})
        self.assertEqual(value, 10)

        # when valid yaml values are passed

    def test_property_string1(self):
        property = {
            "name": "stringProperty",
            "property_id": "stringProperty",
            "description": "testing prop",
            "value": "propertyValue",
            "type": "STRING",
            "tags": "test",
            "segment_rules": [
                {
                    "rules": [
                        {
                            "segments": [
                                "keuyclvf"
                            ]
                        }
                    ],
                    "value": "targeted value",
                    "order": 1
                }
            ],
            "collections": [{
                "collection_id": "appcrash"
            }]
        }

        sut = Property(property)
        self.assertEqual(sut.get_property_name(), "stringProperty")
        self.assertEqual(sut.get_property_id(), "stringProperty")
        self.assertEqual(sut.get_value(), "propertyValue")
        self.assertEqual(sut.get_property_data_type(), ConfigurationType.STRING)
        self.assertEqual(sut.get_property_data_format(), "TEXT")
        value = sut.get_current_value("", {"email": "test.dev@tester.com"})
        self.assertEqual(value, None)

    # when valid yaml values are passed
    def test_property_yaml1(self):
        property = {
            "name": "yamlProperty",
            "property_id": "yamlproperty",
            "description": "testing prop",
            "value": "name: student1\n---\nname: student2",
            "type": "STRING",
            "format": "YAML",
            "tags": "test",
            "segment_rules": [
                {
                    "rules": [
                        {
                            "segments": [
                                "keuyclvf"
                            ]
                        }
                    ],
                    "value": "",
                    "order": 1
                }
            ],
            "collections": [{
                "collection_id": "appcrash"
            }]
        }

        sut = Property(property)
        self.assertEqual(sut.get_property_name(), "yamlProperty")
        self.assertEqual(sut.get_property_id(), "yamlproperty")
        self.assertEqual(sut.get_value()[1]['name'], "student2")
        self.assertEqual(sut.get_property_data_type(), ConfigurationType.STRING)
        self.assertEqual(sut.get_property_data_format(), "YAML")
        value = sut.get_current_value("id1", {"email": "test.dev@tester.com"})
        self.assertEqual(value[0]['name'], "student1")

    # when valid json values are passed
    def test_property_json(self):
        property = {
            "name": "jsonProperty",
            "property_id": "jsonProperty",
            "description": "testing prop",
            "value": {
                "key1": "property value"
            },
            "type": "STRING",
            "format": "JSON",
            "tags": "test"
        }

        sut = Property(property)
        self.assertEqual(sut.get_property_name(), "jsonProperty")
        self.assertEqual(sut.get_property_id(), "jsonProperty")
        self.assertEqual(sut.get_value()['key1'], "property value")
        self.assertEqual(sut.get_property_data_type(), ConfigurationType.STRING)
        self.assertEqual(sut.get_property_data_format(), "JSON")


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import responses
from ibm_appconfiguration.configurations.internal.utils.api_manager import APIManager
from ibm_appconfiguration.configurations.internal.utils.url_builder import URLBuilder


base_url = 'https://cloud.ibm.com'


class MyTestCase(unittest.TestCase):

    def setUp(self) -> None:

        self.responses = responses.RequestsMock()
        self.responses.start()
        self.addCleanup(self.responses.stop)
        self.addCleanup(self.responses.reset)

        URLBuilder.init_with_collection_id(collection_id="collection_id",
                                           guid="guid",
                                           environment_id="environment_id",
                                           region="region",
                                           override_server_host=base_url,
                                           apikey="apiekey")
        URLBuilder.set_auth_type(False)
        self.api_manager = APIManager.get_instance()

    def test_get_call(self):
        mock_response = '{ "features": [], "properties": [], "segments": []}'
        url = 'https://region.apprapp.cloud.ibm.com/apprapp/feature/v1/instances/guid/collections/collection_id/config?environment_id=environment_id'
        self.responses.add(responses.GET,
                      url,
                      body=mock_response,
                      content_type='application/json',
                      status=200)

        resp = self.api_manager.prepare_api_request(method="GET", url=URLBuilder.get_config_url())
        self.assertEqual(resp.get_status_code(), 200)

        try:
            response_data = dict(resp.get_result())
            self.assertEqual(len(response_data), 3)
        except Exception as exception:
            self.fail("Issues with API request")

if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Package to perform the connectivity check.
"""

from threading import Timer
import requests
from .url_builder import URLBuilder


class Connectivity:
    """Connectivity checker"""
    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if Connectivity.__instance is None:
            return Connectivity()
        return Connectivity.__instance

    def __init__(self):
        if Connectivity.__instance is not None:
            print("Connectivity class must be initialized using the get_instance() method")
        else:
            self.__listeners = list()
            Connectivity.__instance = self

    def add_connectivity_listener(self, listener):
        """ Listener for the the internet

        Args:
            listener: Listener for the the internet
        """
        if callable(listener) and not self.__listeners.__contains__(listener):
            self.__listeners.append(listener)

    def check_connection(self):
        """Check the connection"""
        url = URLBuilder.get_network_check_url()
        if url:
            self.__check_network(url)
            timer = Timer(30, self.check_connection)
            timer.daemon = True
            timer.start()

    def __check_network(self, url):
        try:
            _ = requests.head(url, timeout=3)
            for listener in self.__listeners:
                listener(True)
        except Exception as _:
            for listener in self.__listeners:
                listener(False)



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides methods that perform metering and usage related operations.
"""
from threading import Lock, Timer
from datetime import datetime
from .api_manager import APIManager
from .logger import Logger
from ..common import config_messages, config_constants


class Metering:
    """Class to send the metering data."""
    __send_interval = 600
    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if Metering.__instance is None:
            return Metering()
        return Metering.__instance

    def __init__(self):
        """ Virtually private constructor. """
        if Metering.__instance is not None:
            raise Exception("Metering " + config_messages.SINGLETON_EXCEPTION)
        self.__metering_url = None
        self.__repeating = True
        self.__lock = Lock()
        self.__metering_feature_data = dict()
        self.__metering_property_data = dict()
        Metering.__instance = self
        self.send_metering()

    def set_repeat_calls(self, repeat):
        """Set the send_metering repeating task

        Args:
            repeat: Bool to set the repeat task.
        """
        self.__repeating = repeat

    def set_metering_url(self, url: str):
        """Set the metering url

        Args:
            url: Url for the metering.
        """
        self.__metering_url = url

    def add_metering(self, guid: str, environment_id: str,
                     collection_id: str, entity_id: str,
                     segment_id: str, feature_id: str = None,
                     property_id: str = None):
        """ Add the Metering values.

        Args:
            guid: GUID of the App Configuration service. Get it from the service credentials section of the dashboard.
            environment_id: Id of the environment created in App Configuration service instance.
            collection_id: Id of the collection created in App Configuration service instance.
            entity_id: Id of the Entity.
            segment_id: Id of the Segment.
            feature_id: Id of the Feature.
            property_id: Id of the Property.
        """

        self.__lock.acquire()
        try:
            time = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
            feature_json = {
                'count': 1,
                'evaluation_time': time
            }

            modify_metering_data = self.__metering_feature_data if property_id is None else self.__metering_property_data
            modify_id = feature_id if property_id is None else property_id

            if guid in modify_metering_data:
                if environment_id in modify_metering_data[guid]:
                    if collection_id in modify_metering_data[guid][environment_id]:
                        if modify_id in modify_metering_data[guid][environment_id][collection_id]:
                            if entity_id in modify_metering_data[guid][environment_id][collection_id][modify_id]:
                                if segment_id in modify_metering_data[guid][environment_id][collection_id][modify_id][entity_id]:
                                    modify_metering_data[guid][environment_id][collection_id][modify_id][entity_id][segment_id]['evaluation_time'] = time
                                    count = modify_metering_data[guid][environment_id][collection_id][modify_id][entity_id][segment_id]['count']
                                    modify_metering_data[guid][environment_id][collection_id][modify_id][entity_id][segment_id]['count'] = count + 1
                                else:
                                    modify_metering_data[guid][environment_id][collection_id][modify_id][entity_id][segment_id] = feature_json
                            else:
                                modify_metering_data[guid][environment_id][collection_id][modify_id][entity_id] = {
                                    segment_id: feature_json
                                }
                        else:
                            modify_metering_data[guid][environment_id][collection_id][modify_id] = {
                                entity_id: {
                                    segment_id: feature_json
                                }
                            }
                    else:
                        modify_metering_data[guid][environment_id][collection_id] = {
                            modify_id: {
                                entity_id: {
                                    segment_id: feature_json
                                }
                            }
                        }
                else:
                    modify_metering_data[guid][environment_id] = {
                        collection_id: {
                            modify_id: {
                                entity_id: {
                                    segment_id: feature_json
                                }
                            }
                        }
                    }
            else:
                modify_metering_data[guid] = {
                    environment_id: {
                        collection_id: {
                            modify_id: {
                                entity_id: {
                                    segment_id: feature_json
                                }
                            }
                        }
                    }
                }
        finally:
            self.__lock.release()

    def __send_to_server(self, guid, data):
        if self.__repeating:
            api_manager = APIManager.get_instance()
            response = api_manager.prepare_api_request(method="POST",
                                                       url="{0}{1}/usage".format(self.__metering_url, guid),
                                                       data=data)
            status_code = response.get_status_code()
            if 200 <= status_code <= 299:
                Logger.debug("Successfully posted metering data")
            else:
                Logger.debug("Failed to send the metering data")
                Logger.debug(response.get_result())

    def __build_request_body(self, send_metering_data: dict, result: dict, main_key: str):

        for guid, guid_map in send_metering_data.items():
            if guid not in result:
                result[guid] = []
            for environment_id, environment_map in guid_map.items():
                for collection_id, collection_map in environment_map.items():
                    collections_map = {
                        'collection_id': collection_id,
                        'environment_id': environment_id,
                        'usages': []
                    }
                    for feature_id, feature_map in collection_map.items():
                        for entity_id, entity_map in feature_map.items():
                            for segment_id, segment_map in entity_map.items():
                                feature_json = {
                                    main_key: feature_id,
                                    'entity_id': None if entity_id == config_constants.DEFAULT_ENTITY_ID else entity_id,
                                    'segment_id': None if segment_id == config_constants.DEFAULT_SEGMENT_ID else segment_id,
                                    'evaluation_time': segment_map['evaluation_time'],
                                    "count": segment_map['count']
                                }
                                collections_map['usages'].append(feature_json)
                    result[guid].append(collections_map)

    def send_metering(self):
        """Send the metering."""
        if self.__repeating:
            timer = Timer(self.__send_interval, self.send_metering)
            timer.daemon = True
            timer.start()

        self.__lock.acquire()
        try:
            send_feature_data = self.__metering_feature_data
            send_property_data = self.__metering_property_data
            self.__metering_feature_data = dict()
            self.__metering_property_data = dict()
        finally:
            self.__lock.release()

        if len(send_feature_data) <= 0 and len(send_property_data) <= 0:
            return None

        result = dict()

        if len(send_feature_data) > 0:
            self.__build_request_body(send_feature_data, result, 'feature_id')
        if len(send_property_data) > 0:
            self.__build_request_body(send_property_data, result, 'property_id')

        for guid, values in result.items():
            for data in values:
                count = len(data['usages'])
                if count > 25:
                    self.__send_split_metering(guid, data, count)
                else:
                    self.__send_to_server(guid=guid, data=data)
        return result

    def __send_split_metering(self, guid: str, data: dict, count: int):
        limit = 0
        while limit <= count:
            collections_map = {
                'collection_id': data['collection_id'],
                'environment_id': data['environment_id'],
                'usages': data['usages'][limit:limit + 25]
            }
            self.__send_to_server(guid=guid, data=collections_map)
            limit += 25



# (C) Copyright IBM Corp. 2021.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from setuptools import setup, find_packages

NAME = "ibm-appconfiguration-python-sdk"
VERSION = "0.2.3"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = [
    "python-dateutil>=2.8,<3.0.0",
    "requests>=2.20,<3.0",
    "websocket-client==0.57.0",
    "ibm-cloud-sdk-core>=3.10.0",
    "pyyaml>=5.4.1",
    "schema>=0.7.5"
]
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=VERSION,
    author="IBM",
    license='Apache 2.0',
    author_email="mdevsrvs@in.ibm.com",
    description="IBM Cloud App Configuration Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/appconfiguration-python-sdk",
    packages=find_packages(),
    install_requires=REQUIRES,
    include_package_data=True,
    keywords=['python', 'ibm_appconfiguration', 'ibm', 'ibm cloud', 'feature_flags'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3"
    ],
    python_requires='>=3.0'
)



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides methods to construct different url used by the SDK.
"""
from typing import Optional
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator, NoAuthAuthenticator, Authenticator
from .validators import Validators
from ..common import config_constants


class URLBuilder:
    """URLBuilder class to create different urls for the library """
    __wsurl = "/wsfeature"
    __path = "/feature/v1/instances/"
    __service = "/apprapp"
    __events = "/events/v1/instances/"
    __config = "config"
    __override_server_host = ''
    __region = ''
    __http_base = ''
    __web_socket_base = ''
    __config_base = ''
    __iam_authenticator = None
    __hasAuth = True
    __network_check_url = 'https://cloud.ibm.com'

    @classmethod
    def init_with_collection_id(cls, collection_id='', region='', guid='', environment_id='', override_server_host='',
                                apikey=''):
        """Initialise the URLBuilder

        Args:
            collection_id: Id of the collection created in App Configuration service instance.
            region: Region name where the service instance is created.
            guid: GUID of the App Configuration service. Get it from the service credentials section of the dashboard
            environment_id: Id of the environment created in App Configuration service instance.
            override_server_host: Use for testing purpose
            apikey: ApiKey of the App Configuration service. Get it from the service credentials section of the dashboard
        """
        if Validators.validate_string(collection_id) \
                and Validators.validate_string(region) \
                and Validators.validate_string(guid) \
                and Validators.validate_string(environment_id):
            cls.__override_server_host = override_server_host
            cls.__region = region
            cls.__web_socket_base = config_constants.DEFAULT_WSS_TYPE
            cls.__http_base = config_constants.DEFAULT_HTTP_TYPE
            if Validators.validate_string(cls.__override_server_host):
                cls.__iam_authenticator = IAMAuthenticator(apikey, url=config_constants.IAM_TEST_URL)
                cls.__http_base = cls.__override_server_host
                cls.__web_socket_base += (cls.__override_server_host.replace("https://", "").replace("http://", ""))
            else:
                cls.__http_base += region
                cls.__http_base += config_constants.DEFAULT_BASE_URL
                cls.__web_socket_base += region
                cls.__web_socket_base += config_constants.DEFAULT_BASE_URL
                cls.__iam_authenticator = IAMAuthenticator(apikey, url=config_constants.IAM_PROD_URL)

        cls.__http_base += '{0}'.format(cls.__service)
        cls.__config_base = '{0}{1}/collections/{2}/{3}?environment_id={4}'.format(cls.__path,
                                                                                   guid,
                                                                                   collection_id,
                                                                                   cls.__config,
                                                                                   environment_id)

        cls.__web_socket_base += "{0}{1}?instance_id={2}&collection_id={3}&environment_id={4}".format(cls.__service,
                                                                                                      cls.__wsurl,
                                                                                                      guid,
                                                                                                      collection_id,
                                                                                                      environment_id)

    @classmethod
    def set_auth_type(cls, has_auth=True):
        """Set the auth type"""
        cls.__hasAuth = has_auth

    @classmethod
    def get_base_url(cls) -> str:
        """Get the config url"""
        return cls.__http_base

    @classmethod
    def get_config_url(cls) -> str:
        """Get the config url"""
        return cls.__config_base

    @classmethod
    def get_web_socket_url(cls) -> str:
        """Get the web-socket url"""
        return cls.__web_socket_base

    @classmethod
    def get_metering_url(cls) -> str:
        """Get the metering URL"""
        # base = config_constants.DEFAULT_HTTP_TYPE + cls.__region + config_constants.DEFAULT_BASE_URL + cls.__service
        # if Validators.validate_string(cls.__override_server_host):
        #     base = cls.__override_server_host + cls.__service
        return '{0}'.format(cls.__events)

    @classmethod
    def get_iam_authenticator(cls) -> Authenticator:
        """Get the IAM Authenticator

        Returns:
            Authenticator object
        """
        if cls.__hasAuth:
            return cls.__iam_authenticator
        return NoAuthAuthenticator()

    @classmethod
    def get_network_check_url(cls) -> Optional[str]:
        """Return the the network check url"""
        if cls.__hasAuth:
            return cls.__network_check_url
        return None



import unittest
import os
import time
from os.path import join, dirname
from dotenv import load_dotenv

from ibm_appconfiguration import AppConfiguration, Feature, Property

DEFAULT_SERVICE_NAME = 'appconfiguration'


class MyTestCase(unittest.TestCase):

    def setUp(self):
        dotenv_path = join(dirname(__file__), '.env')
        load_dotenv(dotenv_path)
        self.guid = os.environ.get("GUID")
        self.apikey = os.environ.get("APIKEY")
        self.region = os.environ.get("REGION")
        self.collection_id = os.environ.get("COLLECTION_ID")
        self.environment_id = os.environ.get("ENVIRONMENT_ID")

        self.sut = AppConfiguration.get_instance()
        self.sut.init(self.region, self.guid, self.apikey)
        self.sut.enable_debug(True)

    def app_configuration_actions(self):
        features_list = self.sut.get_features()
        feature: Feature = self.sut.get_feature("defaultfeature")

        self.assertEqual(len(features_list), 3)
        self.assertEqual(feature.get_feature_id(), "defaultfeature")

        property_list = self.sut.get_properties()
        property_obj = self.sut.get_property("numericproperty")

        self.assertEqual(len(property_list), 1)
        self.assertEqual(property_obj.get_property_id(), "numericproperty")

        entity_id = "developer_entity"
        entity_attributes = {
            'email': 'tommartin@company.dev'
        }
        current_value = feature.get_current_value(entity_id, entity_attributes)
        self.assertEqual(current_value, "Welcome")

        entity_attributes = {
            'email': 'laila@company.test'
        }
        current_value = feature.get_current_value(entity_id, entity_attributes)
        self.assertEqual(current_value, "Hello")

        entity_attributes = {
            'email': 'tommartin@tester.com'
        }
        current_value = property_obj.get_current_value(entity_id, entity_attributes)
        self.assertEqual(current_value, 81)

        entity_attributes = {
            'email': 'laila@company.test'
        }
        current_value = property_obj.get_current_value(entity_id, entity_attributes)
        self.assertEqual(current_value, 25)

    def test_app_configuration_with_file(self):
        FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'user.json')

        self.sut.set_context(self.collection_id, self.environment_id, options={
            'bootstrap_file': FILE,
            'live_config_update_enabled': False
        })
        time.sleep(2.0)
        self.app_configuration_actions()

    def test_app_configuration_online(self):

        try:
            self.sut.set_context(self.collection_id, self.environment_id)
            self.app_configuration_actions()
            unittest.installHandler()
        except Exception as e:
            print(f'Error {e}')
            unittest.installHandler()


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module defines the model of a segment rule defined in App Configuration service.
"""

class SegmentRules:
    """
       Attributes:
        segment_rules (dict): segment_rules JSON object that contains all the SegmentRules
   """

    def __init__(self, segment_rules: {}):

        self.__order = segment_rules.get("order", 1)
        self.__value = segment_rules.get("value", object)
        self.__rules = segment_rules.get("rules", list())

    def get_order(self) -> int:
        """Get the SegmentRule order"""
        return self.__order

    def get_rules(self) -> list:
        """Get the SegmentRule rules list"""
        return self.__rules

    def get_value(self):
        """Get the SegmentRule value"""
        return self.__value



# Copyright 2021 IBM All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines the model of a segment defined in App Configuration service.
"""

from ..internal.utils.logger import Logger
from .rule import Rule


class Segment:
    """
      Attributes:
         segment_rules (dict): segments JSON object that contains all the Segments
   """

    def __init__(self, segments: {}):
        self.__name = segments.get("name", "")
        self.__segment_id = segments.get("segment_id", "")
        self.__rules = segments.get("rules", list())

    def get_name(self) -> str:
        """Get the Segment name"""
        return self.__name

    def get_segment_id(self) -> str:
        """Get the Segment Id"""
        return self.__segment_id

    def get_rules(self) -> list:
        """Get the Segment rules"""
        return self.__rules

    def evaluate_rule(self, entity_attributes: dict) -> bool:
        """Evaluate the Segment rules

        Args:
            entity_attributes: Entity attributes object
        """
        for index in range(0, len(self.__rules)):
            try:
                dict_sec = self.__rules[index]
                rule = Rule(dict_sec)

                if not rule.evaluate_rule(entity_attributes):
                    return False
            except Exception as exception:
                Logger.debug(f'Invalid action in Segment class, {exception}')
        return True
